package planet.commonapi.factory;

import planet.commonapi.*;
import planet.commonapi.exception.InitializationException;

/**
 * This interface is focused to allow the ability of build networks
 * of different topologies, size and Nodes. The future implementation must
 * read the properties file to known the default
 * values of the future networks. In that are four values:
 * 
 * <ul>
 * <li><b>DEFAULT_NODEFACTORY</b>: The default NodeFactory to use to generate
 * new nodes in the network.</li>
 * <li><b>DEFAULT_SIZE:</b>: The default size of the future networks.</li>
 * <li><b>DEFAULT_K</b>: The default number of bits of the Id that identifies
 * the nodes in the network.</li>
 * <li><b>DEFAULT_TOPOLOGY</b>: The default topology of the network. Its own values
 * are specified by Topology class (or an inherit class for it).</li>
 * <li><b>DEFAULT_SERIALIZEDFILE</b>: The file where appear the serialized state
 * of some network. Only use for the topology Topology.SERIALIZED</b>
 * </ul>
 * 
 * @author Jordi Pujol
 * @see planet.generic.commonapi.factory.Topology Topology
 * @see planet.commonapi.factory.NodeFactory NodeFactory
 * @see planet.commonapi.Network Network
 * @see planet.commonapi.Id Id
 */
public interface NetworkFactory {
	
	/**
	 * Generates a new Network with the actual size of the Network. The class
	 * of Nodes is specified by the actual NodeFactory.
	 * @return A new instance of Network with the number of nodes specified.
	 */
	public Network buildNetwork()
		throws InitializationException;
	
	/**
	 * Generates a new Network with the <b>size</b> of Nodes specified
	 * by parameter. The class of Nodes is specified by the actual NodeFactory.
	 * @param size Number of Nodes to generate in the network.
	 * @return A new instance of Network with the specified <b>size</b>.
	 */
	public Network buildNetwork(int size)
		throws InitializationException;
	
	/**
	 * Generates a new Network with the <b>size</b> of Nodes specified
	 * by parameter. The class of Nodes is specified by <b>nodeFactory</b>.
	 * @param size Number of Nodes to generate in the network.
	 * @param nodeFactory NodeFactory to use to build Nodes for the new network.
	 * @return A new instance of Network with the specified <b>size</b>.
	 */
	public Network buildNetwork(int size, NodeFactory nodeFactory)
		throws InitializationException;
	
	/**
	 * Generates a new Network with the <b>size</b> of Nodes specified
	 * by parameter. The class of Nodes is the obtained by the default nodeFactory.
	 * The <b>topology</b> specifies the network topology for the new network. 
	 * @param size Number of Nodes to generate in the network.
	 * @param topology The network topology for the new network.
	 * @return A new instance of Network with the specified <b>size</b> with
	 * the specified <b>topology</b>.
	 */
	public Network buildNetwork(int size, String topology)
		throws InitializationException;
	
	/**
	 * Generates a new Network with the <b>size</b> of Nodes specified
	 * by parameter. The class of Nodes is specified by <b>nodeFactory</b>.
	 * The <b>topology</b> specifies the network topology for the new network.
	 * @param size Number of Nodes to generate in the network.
	 * @param nodeFactory NodeFactory to use to build Nodes for the new network.
	 * @param topology The network topology for the new network.
	 * @return A new instance of Network with the specified <b>size</b> with
	 * the specified <b>topology</b>.
	 */
	public Network buildNetwork(int size, NodeFactory nodeFactory, String topology)
		throws InitializationException;
	
	/**
	 * Generates a new Network with the <b>size</b> of Nodes specified
	 * by parameter. The class of Nodes is specified by <b>nodeFactory</b>.
	 * The <b>topology</b> specifies the network topology for the new network.
	 * @param size Number of Nodes to generate in the network.
	 * @param nodeFactory NodeFactory to use to build Nodes for the new network.
	 * @param topology The network topology for the new network.
	 * @param k The number of bits for the NodeImpl's Ids.
	 * @return A new instance of Network with the specified <b>size</b> with
	 * the specified <b>topology</b>.
	 * @see planet.commonapi.Id Id
	 */
	public Network buildNetwork(int size, NodeFactory nodeFactory, String topology, int k)
		throws InitializationException;
	
	/**
	 * Returns the actual NodeFactory to use to generate networks.
	 * @return The actual NodeFactory to generate networks.
	 */
	public NodeFactory getNodeFactory();
	
	/**
	 * Sets the NodeFactory to use in the future to generate new networks.
	 * @param nodeFactory NodeFactory's instance to generate new networks.
	 * @see planet.commonapi.factory.NodeFactory NodeFactory
	 * @see Network
	 */
	public void setNodeFactory(NodeFactory nodeFactory);
	
	/**
	 * Returns the actual size of the new Networks.
	 * @return Size of new Networks.
	 * @see Network
	 */
	public int getNetworkSize();
	
	/**
	 * Sets the size of new networks.
	 * @param size Size of new networks.
	 */
	public void setNetworkSize(int size);
	
	/**
	 * Gets the actual network's topology of the generated networks.
	 * @return Topology of the new networks.
	 */
	public String getTopology();
	
	/**
	 * Sets the topology of the new generated networks.
	 * @param topology Topology of the new networks.
	 * @throws InitializationException if the topology is not suported.
	 */
	public void setTopology(String topology) throws InitializationException;
	
	/**
	 * Gets the actual file from where capture a serialized state
	 * of a network. 
	 * @return The filename for the serialized state.
	 */
	public String getSerializedFile();
	
	/**
	 * Sets the filename from where capture a serialized state
	 * of a network.
	 * @param file Filename from read a serialized state of a network.
	 * @throws InitializationException if occur some problem with the filename.
	 */
	public void setSerializedFile(String file) throws InitializationException;
}
