package planet.symphony;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import planet.behaviour.Pattern;
import planet.commonapi.Id;
import planet.commonapi.Message;
import planet.commonapi.NodeHandle;
import planet.commonapi.RouteMessage;
import planet.commonapi.GML.GMLConstraint;
import planet.commonapi.exception.InitializationException;
import planet.generic.commonapi.factory.GenericFactory;
import planet.simulate.Globals;
import planet.simulate.Logger;
import planet.simulate.MessagePool;
import planet.simulate.Results;
import planet.symphony.IdComparer;
import planet.symphony.listeners.LookupListener;
import planet.symphony.messages.JoinMessage;
import planet.symphony.messages.NeighbourMessage;
import planet.util.Properties;
import planet.util.QueueFull;
import planet.util.timer.TimerTaskImpl;

/**
 * 
 * This node is an implementation of the Symphony overlay. 
 * 
 * @author <a href="mailto:heliodoro.tejedor@estudiants.urv.es">Helio Tejedor</a>
 * @author <a href="mailto:marc.sanchez@estudiants.urv.es">Marc Sanchez</a>
 * @author <a href="mailto:Ruben.Mondejar@estudiants.urv.es">Ruben Mondejar</a>
 * @author <a href="mailto:cpairot@etse.urv.es">Carles Pairot</a>
 * @author <a href="mailto:jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 */

public class SymphonyNode
	extends planet.generic.commonapi.NodeImpl
{
	
	/* ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/**
	 * Mode value: Defines the start of communication that requires 
	 * response (REPLY). 
	 */
	public final static int REQUEST					= 0;
	/**
	 * Mode value: Defines the response of a communication.
	 */
	public final static int REPLY					= 1;
	/**
	 * Mode value: Defines a message's mode that requires only
	 * communication on one way.
	 */
	public final static int REFRESH  				= 2;		
	
	/* END ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	/**
	 * Type value: First message that is send by any node to get its
	 * inmediate successor.
	 */
	 public final static int QUERY_JOIN 			 = 0;
     public final static int SET_INFO   			 = 1;
     public final static int QUERY_CONNECT 			 = 2;
     public final static int ACCEPT_CONNECT 		 = 3;
     public final static int CANCEL_CONNECT 		 = 4;
     public final static int CLOSE_LONG_CONNECT 	 = 5;
     public final static int CLOSE_NEIGHBOUR_CONNECT = 6;

     public final static int NumberOfTypes = 8;
     public final static int NumberOfModes = 3;
	/**
	 * Type value: It identifies that this message contains an application 
	 * level Message. 
	 */
	public final static int DATA 			= 7;	
	
	/* END ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	/**
	 * This String contains a string representation of each type
	 * value of the RouteMessage.
	 */
	public final static String[] TYPES = {
			"QUERY_JOIN", "SET_INFO", 
			"QUERY_CONNECT", "ACCEPT_CONNECT", 
			"CANCEL_CONNECT", "CLOSE_LONG_CONNECT", 
			"CLOSE_NEIGHBOUR_CONNECT", "DATA"};
	/**
	 * This String contains a string representation of each mode
	 * value of the RouteMessage.
	 */
	public final static String[] MODES = {"REQUEST","REPLY","REFRESH" };
	
	/* END ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	
	
	
	/**
	 * To build new Id to find a new long distance.
	 */
	private static Random r = new Random();

	
	private Set incommingSet;
	private Set outcommingSet;
	private SortedKList neighbourSet; // with order !!!
	
	private double n;
	
	/* UpdateOutcommingList */
	private Set newSet;
	public int retriesNewLongDistance;
	
	public boolean alive;
	public boolean stabilized;
    private planet.behaviour.BehavioursPool behPool; 
	
	public SymphonyNode (Id id) throws InitializationException {
		super (id);
		endpoints = new Hashtable();
		alive = true;
		
		outcommingSet = new HashSet();
		incommingSet = new HashSet();
		neighbourSet = new SortedKList(new IdComparer(this.nodeHandle), this.nodeHandle, ((SymphonyProperties)Properties.overlayProps).F);
		
		// Stabilize Timer
		setTimer(new StabilizeTask(),  ((SymphonyProperties)Properties.overlayProps).stabilizeSteps,
				((SymphonyProperties)Properties.overlayProps).stabilizeSteps);
		
		newSet = new HashSet();
		stabilized = true;
		behPool = GenericFactory.buildBehavioursPool();
	}
  
	public Hashtable getEndPoints() {
		return this.endpoints;
	}
	
	public RouteMessage buildMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		NodeHandle nextHop = route(to);
		return buildMessage(from, to, route(to), type, mode, m, key);
	}
	
	public RouteMessage buildMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m, String key) {
		NodeHandle nextHop = route(to);
		return buildMessage(from, to, nextHop, type, mode, m, key);
	}
	
	public RouteMessage buildMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		return buildMessage(from, to, nextHop, type, mode, m, key);
	}
	
	public RouteMessage buildMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m, String key) {
		RouteMessage msg = null;
		try {
		    msg = MessagePool.getMessage(key, from, to, type, mode);
			msg.setNextHopHandle(nextHop);
			msg.setMessage(m);
			planet.results.LinkStateResults.newMessage(msg);
			planet.results.LinkStateResults.updateOutcoming(this.id);
		} catch (InitializationException e) {
			key = null;
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n" + e.getMessage(), Logger.ERROR_LOG);
		}
		return msg;
	}
	
    private String sendMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		NodeHandle nextHop = route(to);
		return sendMessage(from, to, route(to), type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m, String key) {
		NodeHandle nextHop = route(to);
		return sendMessage(from, to, nextHop, type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		return sendMessage(from, to, nextHop, type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m, String key) {
		RouteMessage msg = null;
		try {
			msg = MessagePool.getMessage(key, from, to, type, mode);
			msg.setNextHopHandle(nextHop);
			msg.setMessage(m);
			planet.results.LinkStateResults.newMessage(msg);
			planet.results.LinkStateResults.updateOutcoming(this.id);
			send(msg);
		} catch (QueueFull e) {
			key = null;
			Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
		} catch (InitializationException e) {
			key = null;
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n" + e.getMessage(), Logger.ERROR_LOG);
		}
		return key;
	}
	
	private String resendMessage(RouteMessage msg) {
		return resendMessage(msg, msg.getNextHopHandle());
	}
	
	private String resendMessage(RouteMessage msg, NodeHandle nextHop) {
		String key = null;
		try {
			key = msg.getKey();
			msg.setNextHopHandle(nextHop);
			planet.results.LinkStateResults.newMessage(msg);
			planet.results.LinkStateResults.updateOutcoming(this.id);
			send(msg);
		} catch (QueueFull e) {
			key = null;
			Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
		}
		return key;
	}
	
	public void addToNeighbourSet(java.util.Collection c) {
		if (neighbourSet.addAll(c)) {
			updateOutcommingList(); 
		}
	}
	
	public void addToNeighbourSet(NodeHandle neighbour) {
		if (neighbourSet.add(neighbour)) updateOutcommingList();
	}
	
	public Collection getNeighbourSet() {
		return neighbourSet.getSortedSet();
	}
	
	public boolean neighbourSetContains(NodeHandle o) {
		return neighbourSet.contains(o);
	}
	
	public boolean remove(NodeHandle o) {
		return neighbourSet.remove(o);
	}
	
		
	public NodeHandle getSucc() {
		NodeHandle succ = (NodeHandle) neighbourSet.getFirstSucc();
		if (succ == null) return nodeHandle;
		return succ;
	}
	
	public NodeHandle getPred() {
		NodeHandle pred = (NodeHandle) neighbourSet.getFirstPred();
		if (pred == null) return nodeHandle;
		return pred;
	}
	
	public Set getOutcommingSet() {
		return this.outcommingSet;
	}
	
	public Set getIncommingSet() {
		return this.incommingSet;
	}
	
	public Set getNewSet() {
		return this.newSet;
	}
		
	public void lookup(Id key) {
	    String msgkey = null;
        ;
        try {
            msgkey = sendMessage(nodeHandle, GenericFactory.buildNodeHandle(
                    key, true), Globals.LOOKUP, REQUEST, null);
        } catch (InitializationException e) {
            Logger.log("[" + id + "] cannot build a NodeHandle for the key ["
                    + key + "]", Logger.ERROR_LOG);
        }
        addMessageListener(msgkey, new LookupListener(id, key));
        Results.incTraffic();
	};
	
  	public void join(Id bsId) {
		incommingSet.clear();
		outcommingSet.clear();
		newSet.clear();
		
		if (id.equals(bsId)) {
			n = 1;
			stabilized = true;
		} else {
			n = 0;
			try {
				NodeHandle bootstrap = GenericFactory.buildNodeHandle(bsId,true);
				sendMessage(nodeHandle, nodeHandle, bootstrap, QUERY_JOIN, REFRESH, new JoinMessage(nodeHandle, bootstrap));
			} catch (InitializationException e) {
				Logger.log("["+id+"] Cannot to build a NodeHandle before sending a RouteMessage",Logger.ERROR_LOG);
			}
			stabilized = false;
		}
	}
	
	public void leave() {
		Iterator[] c = { 
				incommingSet.iterator(),
				outcommingSet.iterator(),
				neighbourSet.getNeighbourSet().iterator(),
			};
		
			int[] msgtype = {
				CLOSE_LONG_CONNECT,
				CLOSE_LONG_CONNECT,
				CLOSE_NEIGHBOUR_CONNECT,
			};
			for(int i = 0; i < c.length; i++) {
				Iterator it = c[i];
				while (it.hasNext()) {
					NodeHandle current = (NodeHandle) it.next();
					sendMessage(nodeHandle, current, current, msgtype[i], REFRESH, null);
				}
			}
			this.stabilized = true;
			this.alive = false;
	}
	
	public RouteMessage getNewLongDistance() {
		int retries = 10;
	
		while (true) {
			SymphonyId symId = (SymphonyId) id;
			double x = symId.getDoubleValue() + Math.exp(Math.log((double) n) * (r.nextDouble() - 1.0));
			int i = (int) x;
			x -= i;
			Id xid = new SymphonyId(x);
			NodeHandle succ = getSucc();
			NodeHandle pred = getPred();
			if (!xid.between(pred.getId(), id) && !xid.between(id, succ.getId())) 
				try {
					return buildMessage(nodeHandle, GenericFactory.buildNodeHandle(xid, true), nodeHandle, 
						QUERY_CONNECT, REFRESH, null);
				} catch(InitializationException e) {
					Logger.log("[" + id + "] cannot build a NodeHandle for a RouteMessage", Logger.ERROR_LOG);
				}
			else if (retries > 1) retries--;
			else {
					Logger.log(id + ": Cannot find a valid Id to connect it", Logger.MSG_LOG);
					stabilized = true;
					return null;
				}
		}
	}

	private void updateOutcommingList() {
		double nnew = estimation();
	
		boolean recalc = true;
		if ( (outcommingSet.size() > 0) && (n != 0) ) {
			double ratio = (double) nnew / (double) n;
			if ( (ratio >= 0.5) && (ratio <= 2) ) {
				recalc = false;
			}
		}
		if (recalc) {
			n = nnew;
			retriesNewLongDistance = 0;
			stabilized = false;
			RouteMessage newLink = getNewLongDistance();
			if (newLink == null) return;
			try {
				send(newLink);
			} catch(QueueFull e) {
				Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
				MessagePool.freeMessage(newLink);
			}
		}
	}
	
	private double estimation() {
		double length = 0.0;
		int cnt = 0;

		Iterator it = neighbourSet.getNeighbourSet().iterator();
		for(;it.hasNext(); cnt++) {
			NodeHandle neighbour = (NodeHandle) it.next();
			length += ((SymphonyId) id.subtract(neighbour.getId())).getDoubleValue();
		}
		if (cnt == 0) return 1.0;
		else return (double) cnt / length;
	}

	public void dispatcher(RouteMessage msg) {
		/* Broadcast still unavailable */
		NodeHandle src = msg.getSource();
		NodeHandle dest = msg.getDestination();
		
		switch(msg.getMode()) {
		    // Mode = REQUEST, Mode = REQUEST
		  	case REQUEST: case REFRESH: Logger.log("Uncatched message Type [" + msg.getType() + "]", Logger.ERROR_LOG);
		    // Mode = ERROR
		  	case Globals.ERROR:
				switch(msg.getType()) {
					case QUERY_CONNECT:
						RouteMessage newLink = getNewLongDistance();
						if (newLink == null) return;
						else try {
							send(newLink);
						} catch(QueueFull e) {
							Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(newLink);
						}
						break;
					case ACCEPT_CONNECT:
						NodeHandle srcHandle = msg.getSource();
						incommingSet.remove(srcHandle);
						break;
					default:
						Logger.log("Uncatched message Mode [" + msg.getMode() + "]", Logger.ERROR_LOG);
				}
				break;
		  	default: Logger.log("Uncatched message Mode [" + msg.getMode() + "]", Logger.ERROR_LOG);
		}
	}

	public boolean process(int actualStep) {
		super.process(actualStep);
		while (hasMoreMessages()) {
			planet.results.LinkStateResults.updateIncoming(this.id);
			RouteMessage msg = nextMessage();
			try {
				 RouteMessage[] replySet = behPool.onMessage(msg, this);
				 if (replySet == null) return !IsStabilized();
				 else for(int i = 0; i < replySet.length; i++) {
				 		try {
				 			send(replySet[i]);
				 		} catch (QueueFull e) {
				 			Logger.log("Outgoing Queue of Node " + this.id + " is Full", Logger.ERROR_LOG);
				 			MessagePool.freeMessage(msg);
				 		}
				 	}
			} catch (planet.behaviour.exception.NoSuchBehaviourException e) {
				dispatcher(msg);
			} catch (planet.behaviour.exception.NoBehaviourDispatchedException d) {
				dispatcher(msg);
			}
		}
		invokeByStepToAllApplications();
		return !IsStabilized(); 
	}
	
	/**
	 * The isLocalMessage's method is an extension method for commonapi specs. This method 
	 * is used to allow BehavioursPool decide wether the incoming RouteMessage  is for the
	 * local node or for a  remote node. Remenber, this decision may only be addressed  by 
	 * the underlying overlay protocol. For example, for Symphony's Lookup protocol a node 
	 * is  responsible for all  RouteMessages whose keys have as an immediate succesor the 
	 * node's id or have as destination the own node.  
	 * @see planet.behaviour.BehavioursPool
	 * @return True if the incoming RouteMessage taken as input its for the local node.
	 */
	public boolean isLocalMessage(RouteMessage msg) {
		NodeHandle recipient = msg.getDestination();
		return  recipient.equals(this.nodeHandle) 
				|| (getPred() != null && recipient.getId().between(getPred().getId(), this.id));
	}

	public Hashtable getInfo() {
		Hashtable info = new Hashtable();
		info.put("neighbourSet", neighbourSet.getNeighbourSet());
		info.put("estimation", new Double(n));
		return info;
	}
	/**
	 * @param appId String that includes the application identification.
	 * @param to Node that must receive this message. If cannot represents
	 * a real node.
	 * @param nextHop Node that must receive this message as first hop.
	 * @param msg Message to be sent.
	 * If it is null, the message must be routed.
	 */
	public void routeData(String appId,
			NodeHandle to, NodeHandle nextHop, Message msg) {
		NodeHandle hop = nextHop;
		if (hop == null) {
			hop = route(to);
			if (hop == null) Logger.log("route return null!", Logger.ERROR_LOG);
		}
		try {
			sendData(MessagePool.getMessage(GenericFactory.generateKey(),nodeHandle, to, hop, msg, DATA, REQUEST, appId));
		} catch (InitializationException e) {
			Logger.log("["+id+"] cannot build a RouteMessage.",Logger.ERROR_LOG);
		}
	}
	
	public void printNode() {
		StringBuffer sb = new StringBuffer();
		sb.append("________________________________________________");
		sb.append("\nId [" + id + "] >> \n\n");
		sb.append("\nRole>    " + (role?Pattern.Good:Pattern.Bad));
		sb.append("\nN>       " + n);
		Iterator it;
		
		sb.append("\nOutSet>  ");
		it = outcommingSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ",");
		}
		
		sb.append("\nInSet>   ");
		it = incommingSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ",");
		}
		sb.append("\nNeighbourSet> ");
		it = neighbourSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ",");
		}
		sb.append("\n________________________________________________");
		
		
		
		System.out.println(sb.toString());
	}
	
	public void prettyPrintNode() {
	    StringBuffer sb = new StringBuffer();
		sb.append("________________________________________________");
		sb.append("\nId [" + id + "] >> \n\n");
		sb.append("\nRole>    " + (role?Pattern.Good:Pattern.Bad));
		sb.append("\nN>       " + n);
		sb.append("\n________________________________________________");
		System.out.println(sb.toString());
	}
	
	
	public boolean isAlive() {
		return alive;
	}
	
	/* Privates */
	
	private NodeHandle bestOf(NodeHandle to, NodeHandle a, NodeHandle b) {
		if (a == null)
			return b;			
		if (b == null)
			return a;
		double lena = ((SymphonyId) to.getId().subtract(a.getId())).getDoubleValue();
		double lenb = ((SymphonyId) to.getId().subtract(b.getId())).getDoubleValue();
		if (lena <= lenb)
			return a;
		else
			return b;
	}

	private NodeHandle bestOf(NodeHandle to, NodeHandle a, Iterator it) {
		NodeHandle candidate = a;
		
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			candidate = bestOf(to, candidate, current);
		}
		return candidate;
	}
	
	public NodeHandle route(NodeHandle to) {
		if (to.getId().between(id, getSucc().getId())) {
			return getSucc();
		}

		NodeHandle candidate;
		Iterator it;
		
		candidate = null;
		
		it = neighbourSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		it = outcommingSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		/* bidirectional ? */
		it = incommingSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		/* optimize? */
		it = newSet.iterator();
		candidate = bestOf(to, candidate, it);
			
		return candidate;
	}
	
	private void sendData(RouteMessage message) {
		planet.results.LinkStateResults.newMessage(message);
		planet.results.LinkStateResults.updateOutcoming(this.id);
		try {
			send(message);
			Results.incTraffic();
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node "+ id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage (message);
		}
	}

	/* misc */
	public String toString() {
		return "SymphonyNode(" + id + ")";
	}
	
	public boolean IsStabilized() {
		return stabilized;
	}
	
	/**
	 * The networkStabilized's method is a callback method used for the network simulator to
	 * notify to all nodes the stabilization process has finished and the topology is  built
	 * up. 
	 * @param stabilized True wether the network is stable.
	 */
	public void networkStabilized(boolean stabilized) {} 

	private void stabilize() {
		Iterator it = neighbourSet.getNeighbourSet().iterator();
		while(it.hasNext()) {
			NodeHandle neighbour = (NodeHandle) it.next();
			RouteMessage msg = buildMessage(nodeHandle, neighbour, neighbour, SymphonyNode.SET_INFO,
					  SymphonyNode.REFRESH, new NeighbourMessage(neighbourSet.getSortedSet()));
			try {
				send(msg);
			} catch(QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full", Logger.ERROR_LOG);
	 			MessagePool.freeMessage(msg);
			}
		}
	}
	
	/**
	 * Simple TimerTask that invoke stabilize() Node method.
	 * 
	 * @author <a href="mailto: marc.sanchez@estudiants.urv.es">Marc Sanchez</a>
	 *         Date: 07/05/2004
	 */
	public class StabilizeTask extends TimerTaskImpl {
		/**
		 * Initialize this StabilizeTask.
		 */
		public StabilizeTask() {
			super(true);
		}
		/**
		 * Invoke stabilize() method of Node.
		 * 
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			super.run();
			if (!isFinished()) {
				stabilize();
			}
		}
	}
	
	
	
    /**
     * This method throws a NoSuchMethodError, because the broadcast algorithm
     * is not implemented yet.
     * @see planet.commonapi.Node#broadcast(java.lang.String,
     *      planet.commonapi.NodeHandle, planet.commonapi.NodeHandle,
     *      planet.commonapi.Message)
     * @param appId Application identification.
     * @param to Source of the message.
     * @param nextHop NextHop of the message.
     * @param msg Message to be send as broadcast.
     */
	public void broadcast(String appId, NodeHandle to, NodeHandle nextHop,
			Message msg) {
	    throw new NoSuchMethodError("Broadcast is not implemented yet.");
	}
	
    /**
     * The leave protocol is the same as in leave case.
     * @see planet.commonapi.Node#fail()
     *  
     */
	public void fail() {
		leave();
	}
	
    /**
     * Returns up to <b>max</b> successors of the actual node.
     * @see planet.commonapi.Node#getSuccList(int)
     * @param max Maximum number of successors to return.
     * @return Up to <b>max</b> successors.
     */
	public Vector getSuccList(int max) {
        return neighbourSet.getSuccList(max);
	}
	/**
	 * Returns a list of nodes that can be used as next hops on a route
	 * towards key. If is safe, the expected fraction of faulty nodes in the
	 * list is guaranteed to be no higher than the fraction of
	 * faulty nodes in the overlay.
	 * @param key Target key
	 * @param max Number of next hops.
	 * @param safe If true, the expected fraction of faulty nodes are the same
	 * of the nodes in the overlay.
	 * @return Until a maximum of max nodes to use as next hop.
     * @see planet.commonapi.Node#localLookup(planet.commonapi.Id, int, boolean)
     */
    public Vector localLookup(Id key, int max, boolean safe) {
        Vector toReturn = new Vector();
        
        TreeSet contacts = new TreeSet(new IdComparer(nodeHandle));
        contacts.addAll(neighbourSet.getNeighbourSet());
        contacts.addAll(incommingSet);
        contacts.addAll(outcommingSet);

        try {
            Set selected = contacts.headSet(GenericFactory.buildNodeHandle(key,true));
            if (selected.size()<=max)
                toReturn.addAll(selected);
            else 
            {
                Iterator it = selected.iterator();
                for (int i=0; i< max; i++)
                {
                    toReturn.add(it.next());
                }
            }
        } catch (Exception e) {}
        
        return toReturn;
    }
    /**
	 * Obtains an unordered list of up to max nodes that are neighbors 
	 * of the local node.
	 * @param max Maximum of nodes to return.
	 * @return Neighbor nodes.
     * @see planet.commonapi.Node#neighborSet(int)
     */
	public Vector neighborSet(int max) {
	    int max2 = max/2;
        Vector toReturn = neighbourSet.getPredList(max2);
        toReturn.addAll(neighbourSet.getSuccList(max2));
            
        return toReturn;
	}
	/**
     * This methods ALWAYS return false. It is not implemented yet.
     * This operation provides information about ranges of keys for
	 * which the <b>node</b> is currently a root. Returns false if
	 * the range could not be determined, true otherwise.
	 * <br><br>
	 * It is an error to query the range of a node not present in
	 * the neighbor set.
	 * <br><br>
	 * The [leftKey,rightKey] denotes the inclusive range of key values.
	 * @param node Node that is currently a root of some range of keys.
	 * @param rank Number of keys that is root the node (rank=rightKey-leftKey).
	 * @param leftKey The value that appears in the invokation is the candidate left
	 * key of the range. It may be modified to reflect the correct left margin
	 * once invokation has finished.
	 * @param rightKey Shows once the invokation has finished the left margin of
	 * the range.
	 * @return true if the range chold be determined; false otherwise, including
	 * the case of node is not present in the neighbor set returned by neighborSet().
     * @see planet.commonapi.Node#range(planet.commonapi.NodeHandle,
     *      planet.commonapi.Id, planet.commonapi.Id, planet.commonapi.Id)
     */
    public boolean range(NodeHandle node, Id rank, Id leftKey, Id rightKey) {
        return false;
    }
    /**
	 * Returns an ordered set of nodes on which replicas of the object with 
	 * this key can be stored. The maximum number of nodes is maxRank. If
	 * the implementation's maximum replica set size is lower, then its
	 * maximum replica set is returned.
	 * @param key Id for which we find the replica set.
	 * @param maxRank Maximum number of members in replica set.
	 * @return An array of nodes with the replica set.
     * @see planet.commonapi.Node#replicaSet(planet.commonapi.Id, int)
     */
    public Vector replicaSet(Id key, int maxRank) {
        return  neighbourSet.getSuccList(maxRank);
    }
	
	/** 
	 * Returns the NodeHandle closest to <b>id</b>.
	 * @param id Id to find.
	 * @return The NodeHandle closest to <b>id</b>.
	 */
	public NodeHandle getClosestNodeHandle(Id id)
	{
	    try {
	        return route(GenericFactory.buildNodeHandle(id,true));
	    } catch (Exception e)
	    {
	        return (NodeHandle)neighbourSet.getFirstSucc();
	    }
	}
	
	/**
     * Gets all node connections to other nodes as its NodeHandles. The order of
     * NodeHandles is unexpected.
     * 
     * @return A set with all connections as NodeHandles.
     */
    public Set getAllLinks() {
        HashSet conns = new HashSet();

        conns.addAll(outcommingSet);
        conns.addAll(incommingSet);
        conns.addAll(newSet);
        conns.addAll(neighbourSet.getNeighbourSet());

        //remove the null element if exists
        conns.remove(null);
        return conns;
    }
	
	/* BEGIN ************************ GML ****************************/
	/**
	 * This method is a callback method used to collect all of the edges of a graph according to
	 * GML file format for latter visual graph performance.
	 * @param E Edge Set
	 * @param constraint Constraints for edge selection
	 */
	public void getEInGML(Collection E, GMLConstraint constraint) {
		if (E == null || constraint == null) return;
		
		//neighbours (successors and predecessors)
		Iterator it2 = this.neighbourSet.iterator();
		getNewE(it2, constraint, E, "#0000FF");
		
		//long distance links
		Iterator[] it = null;
		it[0] = this.incommingSet.iterator();
		it[1] = this.outcommingSet.iterator();
		for (int i = 0; i < 2; i++) {
			getNewE(it[i],constraint,E,"#FF0000");
		}
	}
	/* END ************************ GML ****************************/
}
