package planet.symphony.behaviours;

import java.util.Iterator;
import java.util.Set;

import planet.behaviour.Behaviour;
import planet.commonapi.Node;
import planet.commonapi.NodeHandle;
import planet.commonapi.RouteMessage;
import planet.symphony.SymphonyNode;
import planet.symphony.SymphonyProperties;
import planet.util.Properties;
/**
 * AcceptConnectBehaviour agrees to accept a long link connection. Source node calls getLongDistanceConnection()
 * method while the outcoming connections ratio is less than K. For that, it sends QUERY_CONNECT messages to the
 * aspirants determined by routing primitive and priori unknown. When an aspirant answers (is immediate successor
 * of the key on QUERY_CONNECT message got by Symphony pdf function), it sends to originator of the QUERY_CONNECT
 * message and ACCEPT_CONNECT message. Then the originator checks if the long peer is valid i.e, is not included
 * on neighbour's set. When enough long valid peers are available they form the new outcommingSet and the long
 * connections of old outcommingSet must be closed because it can be stale and no longer used for originator peer.
 * 
 * @author Marc Sanchez <marc.sanchez@estudiants.urv.es>
 * @author Helio Tejedor <heliodoro.tejedor@estudiants.urv.es>
 */
public class AcceptConnectBehaviour implements Behaviour {
	/**
	 *  Given a RouteMessage and a Node as input, onMessage's method do execution 
	 * steps for the behaviour. Node should be casted to Node's class in order to
	 * manage data structures and protocol of designed overlay.
	 * @param msg RouteMessage taken as input.
	 * @param node Node taken as input.
	 * @return Returns either an array of RouteMessages or null when no messages
	 * need to transmit this node. 
	 */
	public RouteMessage[] onMessage(RouteMessage msg, Node node) {
		SymphonyNode symphony = (SymphonyNode) node;
		
		Set outcommingSet = symphony.getOutcommingSet();
		Set incommingSet  = symphony.getIncommingSet();
		Set newSet = symphony.getNewSet();
		
		/* We suppose long links are bidireccional TCP connections.  Hence,
		 * only one of the two peers must stablish the connection. The peer
		 * with little id removes the peer of incomming set. The other peer
		 * asks for a new connection. This event occurs  when the two peers
		 * want to stablish a long link with each other. 
		 */ 
		NodeHandle longHandle = msg.getSource();
		if (incommingSet.contains(longHandle)) {
			if (symphony.getId().compareTo(longHandle.getId()) < 0) {
				incommingSet.remove(longHandle);
			} else {
				RouteMessage newQueryConnect = symphony.getNewLongDistance();
				if (newQueryConnect == null) return null;
				return new RouteMessage[] { newQueryConnect }; 
			}
		}
		
		boolean alreadyInNewSet = newSet.contains(longHandle);
		/* If connection is not to a member of neighbour's set, we add it to new connection set, outcoming set and
		 * remove it from incoming set. We have stablish new long links with nodes not on  neighbour's set.
		 */
		if (!symphony.neighbourSetContains(longHandle)) {
			newSet.add(longHandle);
			outcommingSet.add(longHandle); 
			incommingSet.remove(longHandle); 
		}
		
		/*
		 * If not in new connection's set, reset retries for new long distances.
		 */
		if (!alreadyInNewSet) symphony.retriesNewLongDistance = 0;
		else { 
			symphony.retriesNewLongDistance++;
			if (symphony.retriesNewLongDistance >= ((SymphonyProperties) Properties.overlayProps).maxRetriesNewLongDistance) {
				symphony.stabilized = true;
				return null;
			}
		}
		
		/*
		 *  If enough long link's apirants, the new outcommingSet is the new  long distances set
		 *  and the old outcomming's Set is now stale. Therefore, the peers must close the stale 
		 *  long connections. Otherwise, we call getNewLongDistance() method.
		 */
		if (newSet.size() == ((SymphonyProperties)Properties.overlayProps).K) {
			Set staleSet  = outcommingSet;
			outcommingSet = newSet;
			// Now node is stable
			symphony.stabilized = true;
			
			RouteMessage[] close = null;
			Iterator it = staleSet.iterator();
			if (it.hasNext()) close = new RouteMessage[staleSet.size()];
			else return null;
			for(int i = 0; it.hasNext(); i++) {
				NodeHandle longPeerWithStaleConnection = (NodeHandle) it.next();
				close[i] = symphony.buildMessage(symphony.getLocalHandle(), longPeerWithStaleConnection, 
						SymphonyNode.CLOSE_LONG_CONNECT, SymphonyNode.REFRESH, null);
			}
			staleSet.clear();
			return close;
		} else { 
			RouteMessage newQueryConnect = symphony.getNewLongDistance();
			if (newQueryConnect == null) return null;
			return new RouteMessage[] { newQueryConnect };
		}
	}
	/**
	 * @return Returns the name of behaviour. 
	 */
	public String getName() {
		return "AcceptConnectBehaviour";
	}
	/**
	 * @return Returns a string representation of the behaviour. In general, the toString
	 * method returns a string that "textually represents" this behaviour. The result should 
	 * be a concise but informative representation that is easy for a person to read. 
	 */ 
	public String toString() {
		return getName();
	}
}
