package planet.symphony.behaviours;

import planet.behaviour.Behaviour;
import planet.commonapi.Node;
import planet.commonapi.NodeHandle;
import planet.commonapi.RouteMessage;
import planet.symphony.SymphonyNode;
import planet.symphony.SymphonyProperties;
import planet.util.Properties;

import java.util.Set;

/**
 * QueryConnectBehaviour determines if the QUERY_CONNECT issuer can connect or not  to a 
 * remote node. This Behaviour is provided in order to stablish long links between nodes. 
 * The behaviour checks if already holds a connection from the incoming  node looking up
 * to outcoming, incoming  and neighbour sets. Otherwise, it  accepts the new connection 
 * only if the incoming connection's rate is fewer than K.     
 * 
 * @author Marc Sanchez <marc.sanchez@estudiants.urv.es>
 * @author Helio Tejedor <heliodoro.tejedor@estudiants.urv.es>
 */
public class QueryConnectBehaviour implements Behaviour {
	/**
	 *  Given a RouteMessage and a Node as input, onMessage's method do execution 
	 * steps for the behaviour. Node should be casted to Node's class in order to
	 * manage data structures and protocol of designed overlay.
	 * @param msg RouteMessage taken as input.
	 * @param node Node taken as input.
	 * @return Returns either an array of RouteMessages or null when no messages
	 * need to transmit this node. 
	 */
	public RouteMessage[] onMessage(RouteMessage msg, Node node) {
		SymphonyNode symphony = (SymphonyNode) node;
		
		NodeHandle sourceHandle = msg.getSource();
		String key = msg.getKey();
		
		Set outcommingSet = symphony.getOutcommingSet();
		Set incommingSet  = symphony.getIncommingSet();
		Set newSet = symphony.getNewSet();
		
		boolean accepted;
		if (symphony.neighbourSetContains(sourceHandle) 
				|| newSet.contains(sourceHandle)
				|| outcommingSet.contains(sourceHandle)) {
			accepted = false;
		} else if (incommingSet.contains(sourceHandle)) {
			accepted = true;
		} else if (incommingSet.size() < ((SymphonyProperties)Properties.overlayProps).K) {
			incommingSet.add(sourceHandle);
			accepted = true;
		} else accepted = false;
		
		msg.setType(accepted?SymphonyNode.ACCEPT_CONNECT : SymphonyNode.CANCEL_CONNECT );
		msg.setDestination(msg.getSource());
		msg.setSource(symphony.getLocalHandle());
		planet.results.LinkStateResults.newMessage(msg);
		planet.results.LinkStateResults.updateOutcoming(symphony.getId());
		return new RouteMessage[] { msg };
	}
	/**
	 * @return Returns the name of behaviour. 
	 */
	public String getName() {
		return "QueryConnectBehaviour";
	}
	/**
	 * @return Returns a string representation of the behaviour. In general, the toString
	 * method returns a string that "textually represents" this behaviour. The result should 
	 * be a concise but informative representation that is easy for a person to read. 
	 */ 
	public String toString() {
		return getName();
	}
}
